"""Notebook namespaces."""

import ast
from ast import NodeVisitor
from collections import defaultdict
from functools import partial
from inspect import getsource
from textwrap import dedent
from types import SimpleNamespace

from cachier import cachier
from nbformat import NO_CONVERT, reads
from ploomber_engine._util import parametrize_notebook
from ploomber_engine.ipython import PloomberClient

from boilercore.hashes import hash_args
from boilercore.notebooks.types import Attributes, Params, SimpleNamespaceReceiver

NO_ATTRS = []
NO_PARAMS = {}


def get_ns_attrs(receiver: SimpleNamespaceReceiver) -> list[str]:
    """Get the list of attribute accesses in the `ns` namespace in the receiver."""
    attributes = AccessedAttributesVisitor()
    attributes.visit(ast.parse(dedent(getsource(receiver))))
    return list(attributes.names["ns"])


def get_nb_ns(
    nb: str, params: Params = NO_PARAMS, attributes: Attributes = NO_ATTRS
) -> SimpleNamespace:
    """Get notebook namespace, optionally parametrizing or limiting returned attributes.

    Args:
        nb: Notebook contents as text.
        params: Parameters to inject below the first `parameters`-tagged code cell.
        attributes: If given, limit the notebook attributes to return in the namespace.

    Returns
    -------
        Notebook namespace.

    Example:
        ```Python
        ns = get_nb_ns(
            nb=Path("notebook.ipynb").read_text(encoding="utf-8"),
            params={"some_parameter": 3, "other_parameter": [0, 1]},
            attributes=["results", "other_attribute_to_return"],
        )
        ```
    """
    nb_client = get_nb_client(nb)
    # We can't just `nb_client.execute(params=...)` since`nb_client.get_namespace()`
    # would execute all over again
    if params:
        parametrize_notebook(nb_client._nb, parameters=params)  # noqa: SLF001
    namespace = nb_client.get_namespace()
    return SimpleNamespace(
        **(
            {
                attr: namespace[attr]
                for attr in attributes
                if namespace.get(attr) is not None
            }
            if attributes
            else namespace
        )
    )


@cachier(hash_func=partial(hash_args, get_nb_ns))
def get_cached_nb_ns(
    nb: str, params: Params = NO_PARAMS, attributes=NO_ATTRS
) -> SimpleNamespace:
    """Get cached notebook namespace, optionally parametrizing or limiting attributes.

    This caches the return values and avoids execution if the hash of input argument
    values matches an earlier call.

    Args:
        nb: Notebook contents as text.
        params: Parameters to inject below the first `parameters`-tagged code cell.
        attributes: If given, limit the notebook attributes to return in the namespace.

    Returns
    -------
        Notebook namespace.

    Example:
        ```Python
        ns = get_cached_nb_ns(
            nb=Path("notebook.ipynb").read_text(encoding="utf-8"),
            params={"some_parameter": 3, "other_parameter": [0, 1]},
            attributes=["results", "other_attribute_to_return"],
        )
        ```
    """
    return get_nb_ns(nb, params, attributes)


def get_nb_client(nb: str) -> PloomberClient:
    """Get notebook client."""
    return PloomberClient(reads(nb, as_version=NO_CONVERT))


class AccessedAttributesVisitor(NodeVisitor):
    """Map accessed attributes to their namespaces."""

    def __init__(self):
        self.names: dict[str, set[str]] = defaultdict(set)

    def visit_Attribute(self, node: ast.Attribute):  # noqa: N802, D102
        if isinstance(node.value, ast.Name) and not node.attr.startswith("__"):
            self.names[node.value.id].add(node.attr)
        self.generic_visit(node)



"""Paths and modules."""

from collections.abc import Iterable
from contextlib import closing
from dataclasses import dataclass
from datetime import datetime
from importlib.machinery import ModuleSpec
from os import walk
from pathlib import Path
from re import NOFLAG, VERBOSE, Match, compile
from shlex import quote
from string import Template
from types import ModuleType

from dulwich.porcelain import status, submodule_list
from dulwich.repo import Repo


def get_package_dir(package: ModuleType) -> Path:
    """Get the directory of a package given the top-level module."""
    return Path(package.__spec__.submodule_search_locations[0])  # type: ignore


def get_module_name(module: ModuleType | ModuleSpec | Path | str) -> str:
    """Get an unqualified module name.

    Example: `get_module_name(__spec__ or __file__)`.
    """
    if isinstance(module, ModuleType | ModuleSpec):
        return get_qualified_module_name(module).split(".")[-1]
    path = Path(module)
    return path.parent.name if path.stem in ("__init__", "__main__") else path.stem


def get_qualified_module_name(module: ModuleType | ModuleSpec) -> str:  # type: ignore
    """Get a fully-qualified module name.

    Example: `get_module_name(__spec__ or __file__)`.
    """
    if isinstance(module, ModuleType):
        module: ModuleSpec = module.__spec__  # type: ignore
    return module.name


def dt_fromisolike(match: Match[str], century: int | str = 20) -> datetime:
    """Get datetime like ISO 8601 but with flexible delimeters and missing century."""
    m: dict[str, str] = {
        **{grp: val or "00" for grp, val in match.groupdict().items()},
        **{grp: match.group(grp) for grp in ("century", "tz", "sym")},
    }
    tz = m["tz"] or ""
    if tz and tz.casefold() != "z":
        tz = (
            f"{m['sym']}{m['tz_hour']}:{m['tz_minute']}:{m['tz_second']}"
            f".{m['tz_fraction']}"
        )
    return datetime.fromisoformat(
        Template("$year-$month-${day}T$hour:$minute:$second.$fraction$tz").substitute(
            year=f"{m['century'] or str(century)}{m['decade']}",
            tz=tz,
            **{
                name: m[name]
                for name in ["month", "day", "hour", "minute", "second", "fraction"]
            },
        )
    )


GROUP = Template(r"(?P<$name>\d{$n})")
SUBSTITUTIONS = {
    "D": r"[^TtZz+\d]",  # A valid digit delimeter is not T, Z, +, or a digit
    **{
        # Fractional seconds have at least one digit
        name: GROUP.substitute(name=name, n="1,")
        for name in ["fraction", "tz_fraction"]
    },
    **{
        # All other groups have exactly two digits
        name: GROUP.substitute(name=name, n="2")
        for name in [
            "century",
            "decade",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "tz_hour",
            "tz_minute",
            "tz_second",
        ]
    },
}
ISOLIKE_PATTERN = Template(
    r"""
        $century?$decade  # Century is optional
        $D$month
        $D$day
        (?:  # Time is optional
            [Tt]
            $hour  # Only the hour is required
            (?:$D$minute)?
            (?:$D$second)?
            (?:$D$fraction)?
            (?P<tz>  # Timezone information is optional
                [Zz]  # It can be Z or z
                |(?:  # Or it can be a delta
                    (?P<sym>[\+-])$tz_hour  # Only the delta hour is required
                    (?:$D$tz_minute)?
                    (?:$D$tz_second)?
                    (?:$D$tz_fraction)?
                )
            )?
        )?
    """
).substitute(SUBSTITUTIONS)
ISOLIKE = compile(flags=VERBOSE, pattern=ISOLIKE_PATTERN)

DEFAULT_SUFFIXES = [".py"]


def map_stages(package: Path, suffixes=DEFAULT_SUFFIXES) -> dict[str, Path]:
    """Map stage module names to their paths."""
    modules: dict[str, Path] = {}
    for path in walk_module_paths(package, suffixes):
        module = get_module_rel(
            get_qualified_module_name_from_paths(path, package), package.name
        )
        if module == ".":
            continue
        modules[module.replace(".", "_")] = path
    return modules


def get_module_rel(module: str, relative: str) -> str:
    """Get module name relative to another module."""
    if relative not in module:
        raise ValueError(f"{module} not relative to {relative}.")
    return (
        "."
        if module == relative
        else compile(rf"^.*{relative}\.").sub(repl="", string=module)
    )


def walk_modules(
    package: Path, suffixes: list[str] = DEFAULT_SUFFIXES
) -> Iterable[str]:
    """Walk modules from a given submodule path and the top level library directory."""
    for module, _ in walk_module_map(package, suffixes=suffixes):
        yield module


def walk_module_map(
    package: Path, suffixes: list[str] = DEFAULT_SUFFIXES
) -> Iterable[tuple[str, Path]]:
    """Walk the map of modules to their paths."""
    for module in walk_module_paths(package, suffixes=suffixes):
        yield get_qualified_module_name_from_paths(module, package), module


def get_qualified_module_name_from_paths(module: Path, package: Path) -> str:
    """Get the qualified name of a module file relative to a package file."""
    module = module.parent if module.stem in ("__init__", "__main__") else module
    return (
        str(module.relative_to(package.parent).with_suffix(""))
        .replace("\\", ".")
        .replace("/", ".")
    )


def walk_module_paths(
    package: Path, suffixes: list[str] = DEFAULT_SUFFIXES
) -> Iterable[Path]:
    """Walk the paths of a Python package."""
    yield from (
        match
        for match in walk_matches(
            package,
            regex=rf"""^
                 [^\.]*  # Match file stem
                 {get_suffixes_re(suffixes)}  # Match suffixes
                 $""",
            root_regex=r"^(?!__).*$",  # Exclude dunder folders
            flags=VERBOSE,
        )
        if match.stem != "__init__"
    )


def get_suffixes_re(suffixes: list[str]) -> str:
    """Get the regex pattern string for file suffixes."""
    suffixes_re = f"({get_suffix_re(suffixes[0])}"
    for suffix in suffixes[1:]:
        suffixes_re += f"|{get_suffix_re(suffix)}"
    suffixes_re += ")"
    return suffixes_re


def get_suffix_re(suffix: str) -> str:
    """Get the regex pattern string for a file suffix."""
    suffix = suffix.replace(r"\.", ".")
    if not suffix.startswith("."):
        suffix = f".{suffix}"
    return suffix.replace(".", r"\.")


def walk_matches(
    path: Path,
    glob: str | None = None,
    regex: str | None = None,
    root_regex: str | None = None,
    flags=NOFLAG,
) -> Iterable[Path]:
    """Walk a directory returning regex or glob matches."""
    glob = glob or "*"
    path_re = compile(regex or "^.*$", flags=flags)
    root_re = compile(root_regex or "^.*$", flags=flags)
    for root, _dirs, files in walk(path):
        if not root_re.match(Path(root).name):
            continue
        files = [Path(root) / file for file in files]
        for path in sorted(Path(root).glob(glob)):
            if path in files and path_re.match(path.name):
                yield path


def fold(path: Path, resolve: bool = True) -> str:
    """Resolve and normalize a path to a POSIX string path with forward slashes."""
    return quote(str(path.resolve() if resolve else path).replace("\\", "/"))


def modified(nb: Path | str) -> bool:
    """Check whether notebook is modified."""
    return Path(nb) in (change.resolve() for change in get_changes())


def get_changes() -> list[Path]:
    """Get pending changes."""
    staged, unstaged, _ = status(untracked_files="no")
    changes = {
        # Many dulwich functions return bytes for legacy reasons
        Path(path.decode("utf-8")) if isinstance(path, bytes) else path
        for change in (*staged.values(), unstaged)
        for path in change
    }
    # Exclude submodules from the changeset (submodules are considered always changed)
    return sorted(
        change
        for change in changes
        if change not in {submodule.path for submodule in get_submodules()}
    )


@dataclass
class Submodule:
    """Represents a git submodule."""

    _path: str | bytes
    """Submodule path as reported by the submodule source."""
    commit: str
    """Commit hash currently tracked by the submodule."""
    path: Path = Path()
    """Submodule path."""
    name: str = ""
    """Submodule name."""

    def __post_init__(self):
        """Handle byte strings reported by some submodule sources, like dulwich."""
        # Many dulwich functions return bytes for legacy reasons
        self.path = Path(
            self._path.decode("utf-8") if isinstance(self._path, bytes) else self._path
        )
        self.name = self.path.name


def get_submodules() -> list[Submodule]:
    """Get the special template and typings submodules, as well as the rest."""
    with closing(repo := Repo(str(Path.cwd()))):
        return [Submodule(*item) for item in list(submodule_list(repo))]



"""Formatting and display utilities."""

from asyncio import create_subprocess_exec
from asyncio.subprocess import PIPE
from shlex import split
from subprocess import CalledProcessError
from typing import Any

import pandas as pd
from IPython.core.display import Markdown, Math
from IPython.display import display
from sympy import FiniteSet
from sympy.printing.latex import latex


def set_format():
    """Set up formatting for interactive notebook sessions.

    The triple curly braces in the f-string allows the format function to be dynamically
    specified by a given float specification. The intent is clearer this way, and may be
    extended in the future by making `float_spec` a parameter.
    """
    float_spec = ":#.4g"
    pd.options.display.min_rows = pd.options.display.max_rows = 50
    pd.options.display.float_format = f"{{{float_spec}}}".format


def disp_named(*args: tuple[Any, str]):
    """Display objects with names above them."""
    for elem, name in args:
        display(Markdown(f"##### {name}"))
        display(elem)


def disp_free(title, eqn, **kwargs):
    """Display free symbols."""
    disp(title, eqn, **kwargs)
    disp("Free symbols", FiniteSet(*eqn.rhs.free_symbols), **kwargs)


def disp(title, *exprs, **kwargs):
    """Display equation."""
    print(f"{title}:")  # noqa: T201
    display(*(math_mod(expr, **kwargs) for expr in exprs))


def math_mod(expr, long_frac_ratio=3, **kwargs):
    """Represent expression as LaTeX math."""
    return Math(latex(expr, long_frac_ratio=long_frac_ratio, **kwargs))


async def clean(nb: str):
    """Clean a notebook."""
    commands = [
        f"ruff --fix-only {nb}",
        f"ruff format {nb}",
         "   nb-clean clean --remove-empty-cells"
         "     --preserve-cell-outputs"
         "     --preserve-cell-metadata special tags"
        f"    -- {nb}",
    ]  # fmt: skip
    for command in commands:
        await run_process(command)


async def run_process(command: str, venv: bool = True) -> str:
    """Run a process asynchronously."""
    command, *args = split(command)
    process = await create_subprocess_exec(
        f"{'.venv/scripts/' if venv else ''}{command}", *args, stdout=PIPE, stderr=PIPE
    )
    stdout, stderr = (msg.decode("utf-8") for msg in await process.communicate())  # type: ignore  # pyright 1.1.347  # Implicit iter
    message = (
        (f"{stdout}\n{stderr}" if stdout and stderr else stdout or stderr)
        .replace("\r\n", "\n")
        .strip()
    )
    if process.returncode:
        exception = CalledProcessError(
            returncode=process.returncode, cmd=command, output=stdout, stderr=stderr
        )
        exception.add_note(message)
        exception.add_note("Arguments:\n" + "    \n".join(args))
        raise exception
    return message



"""Types."""

from pathlib import Path
from typing import Literal, TypeAlias, TypeVar

T = TypeVar("T", bound=Path | str)
PathOrPaths: TypeAlias = T | list[T] | dict[str, T]
Paths: TypeAlias = dict[str, PathOrPaths[T]]
Bound: TypeAlias = float | Literal["-inf", "inf"]



"""Model fits."""

import warnings
from collections.abc import Mapping, Sequence
from functools import partial
from typing import Any, Literal
from warnings import catch_warnings

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from scipy.optimize import OptimizeWarning, curve_fit
from scipy.stats import t
from uncertainties import ufloat

from boilercore.models.fit import Fit
from boilercore.types import Bound, Guess

XY_COLOR = (0.2, 0.2, 0.2)
"""Default color for measurement points."""

CONFIDENCE_INTERVAL_95 = t.interval(0.95, 1)[1]
"""Confidence interval for a single sample from a student's t-distribution."""


def fit_and_plot(
    model: Any,
    params: Fit,
    x: Any,
    y: Any,
    y_errors: Any = None,
    confidence_interval=CONFIDENCE_INTERVAL_95,
    ax: Axes | None = None,
    run: str | None = None,
) -> tuple[dict[str, float], dict[str, float]]:
    """Get fits and errors for project model and plot the results."""
    if not ax:
        _fig, ax = plt.subplots()
    fits, errors = fit_from_params(
        model=model,
        params=params,
        x=x,
        y=y,
        y_errors=y_errors,
        confidence_interval=confidence_interval,
    )
    plot_fit(
        model=model,
        x=x,
        y=y,
        y_0=fits["T_s"],
        params=fits | params.fixed_values,
        errors=errors | dict.fromkeys(params.fixed_errors, 0),
        y_errors=y_errors if y_errors is not None else None,
        ax=ax,
        run=run or "run",
    )
    return fits, errors


def fit_from_params(
    model: Any,
    params: Fit,
    x: Any,
    y: Any,
    y_errors: Any = None,
    confidence_interval: float = CONFIDENCE_INTERVAL_95,
) -> tuple[dict[str, float], dict[str, float]]:
    """Get fits and errors for project model."""
    fits, errors = fit(
        model=model,
        fixed_values=params.fixed_values,
        free_params=params.free_params,
        initial_values=params.initial_values,
        model_bounds=params.model_bounds,
        x=x,
        y=y,
        y_errors=y_errors,
        confidence_interval=confidence_interval,
    )
    return (
        dict(zip(params.free_params, fits, strict=True)),
        dict(zip(params.free_errors, errors, strict=True)),
    )


def fit(
    model: Any,
    fixed_values: Any,
    free_params: list[str],
    initial_values: Mapping[str, Guess],
    model_bounds: Mapping[str, Bound],
    x: Any,
    y: Any,
    y_errors: Any = None,
    confidence_interval: float = CONFIDENCE_INTERVAL_95,
    method: Literal["trf", "dogbox"] = "trf",
) -> tuple[Any, Any]:
    """Get fits and errors."""
    # Perform fit, filling "nan" on failure or when covariance computation fails
    with catch_warnings():
        warnings.simplefilter("error", category=OptimizeWarning)
        try:
            fits, pcov = curve_fit(
                f=partial(model, **fixed_values),
                p0=get_guesses(free_params, initial_values),
                # Expects e.g. ([L1, L2, L3], [H1, H2, H3])
                bounds=tuple(zip(*get_bounds(free_params, model_bounds), strict=True)),
                xdata=x,
                ydata=y,
                sigma=None if y_errors is None else y_errors,
                absolute_sigma=y_errors is not None,
                method=method,
            )
        except (RuntimeError, OptimizeWarning):
            dim = len(free_params)
            fits = np.full(dim, np.nan)
            pcov = np.full((dim, dim), np.nan)

    # Compute confidence interval
    standard_errors = np.sqrt(np.diagonal(pcov))
    errors = standard_errors * confidence_interval

    # Catching `OptimizeWarning` should be enough, but let's explicitly check for inf
    fits = np.where(np.isinf(errors), np.nan, fits)
    errors = np.where(np.isinf(errors), np.nan, errors)
    return fits, errors


def get_guesses(
    params: Sequence[str], guesses: Mapping[str, Guess]
) -> tuple[Guess, ...]:
    """Compose guesses."""
    return tuple(guess for param, guess in guesses.items() if param in params)


def get_bounds(params: Sequence[str], bounds: Mapping[str, Bound]) -> tuple[Bound, ...]:
    """Compose bounds."""
    return tuple(bound for param, bound in bounds.items() if param in params)


def plot_fit(
    model: Any,
    x: Any,
    y: Any,
    y_0: float,
    params: Mapping[str, Any],
    errors: Mapping[str, Any],
    y_errors: Sequence[Any] | None = None,
    ax: Axes | None = None,  # type: ignore
    run: str | None = None,
):
    """Plot a model fit."""
    if not ax:
        _fig, ax = plt.subplots()
        ax: Axes
    ax.margins(0, 0)
    ax.set_title(f"{run = }" if run else "run")
    ax.set_xlabel("x (m)")
    ax.set_ylabel("T (C)")

    # Plot measurements
    ax.plot(x, y, ".", label="Measurements", color=XY_COLOR, markersize=10)
    (xlim_min, xlim_max) = (0, max(x))
    pad = 0.025 * (xlim_max - xlim_min)
    x_smooth = np.linspace(xlim_min - pad, xlim_max + pad, 200)
    y_smooth, y_min, y_max = get_model_with_error(model, x_smooth, params, errors)
    ax.plot(x_smooth, y_smooth, "--", label="Model Fit")

    # Plot measurement errors
    if y_errors is not None:
        ax.errorbar(x=x, y=y, yerr=y_errors, fmt="none", color=XY_COLOR)

    # Plot confidence band
    ax.fill_between(
        x=x_smooth,
        y1=y_min,  # type: ignore  # pyright 1.1.333
        y2=y_max,  # type: ignore  # pyright 1.1.333
        color=[0.8, 0.8, 0.8],
        edgecolor=[1, 1, 1],
        label="95% CI",
    )

    # Extrapolate
    ax.plot(0, y_0, "x", label="Extrapolation", color=[1, 0, 0])

    # Finish
    ax.legend()


def get_model_with_error(model, x, params, errors):
    """Evaluate the model for x and return y with errors."""
    u_x = [ufloat(v, 0, "x") for v in x]
    u_y = model(u_x, **combine_params_and_errors(params, errors))
    y = np.array([v.nominal_value for v in u_y])
    y_min = y - [v.std_dev for v in u_y]
    y_max = y + [v.std_dev for v in u_y]
    return y, y_min, y_max


def combine_params_and_errors(
    params: Mapping[str, Any], errors: Mapping[str, Any]
) -> dict[str, Any]:
    """Return parameters with errors given mappings, one with `_err`-suffixed keys."""
    return dict(
        zip(
            params.keys(),
            [
                ufloat(param, error, tag)
                for param, error, tag in zip(
                    params.values(),
                    {param: errors[f"{param}_err"] for param in params}.values(),
                    params.keys(),
                    strict=True,
                )
            ],
            strict=True,
        )
    )



"""Project parameters."""

from pathlib import Path

from pydantic.v1 import Field

from boilercore import get_params_file
from boilercore.models import SynchronizedPathsYamlModel
from boilercore.models.fit import Fit
from boilercore.models.geometry import Geometry
from boilercore.models.paths import Paths

PARAMS_FILE = get_params_file()


class Params(SynchronizedPathsYamlModel):
    """Global project parameters."""

    fit: Fit = Field(default_factory=Fit, description="Model fit parameters.")
    geometry: Geometry = Field(default_factory=Geometry, description="Geometry.")
    paths: Paths = Field(default_factory=Paths)

    def __init__(self, data_file: Path = PARAMS_FILE, **kwargs):
        super().__init__(data_file, **kwargs)


PARAMS = Params()
"""All project parameters, including paths."""



"""Generate reports for notebooks tracked by DVC."""

from asyncio import TaskGroup
from collections.abc import Callable, Coroutine
from concurrent.futures import ProcessPoolExecutor
from contextlib import AbstractContextManager
from functools import wraps
from os import chdir
from pathlib import Path
from shlex import split
from sys import stdout
from tempfile import NamedTemporaryFile
from typing import Any
from urllib import request

from dulwich.porcelain import add
from dvc.repo import Repo
from loguru import logger
from ploomber_engine import execute_notebook

from boilercore import notebooks
from boilercore.paths import fold

logger.remove()
logger.add(
    sink=stdout, enqueue=True, format=("<green>{time:mm:ss}</green> | {message}")
)
logger = logger.opt(colors=True)

ZOTERO = Path(NamedTemporaryFile().name)
"""Temporary file path which will contain the Zotero Lua filter.

Usage conforms to [MIT license][#license] of `retorquere/zotero-better-bibtex`.

[#license]: <https://raw.githubusercontent.com/retorquere/zotero-better-bibtex/v6.7.164/LICENSE>
"""
ZOTERO.write_bytes(
    request.urlopen(
        "https://raw.githubusercontent.com/retorquere/zotero-better-bibtex/v6.7.164/site/content/exporting/zotero.lua"
    ).read()
)


async def generate(paths, repo: Repo | None = None):
    """Generate reports for notebooks."""
    nbs = get_nbs(repo, paths)
    if not nbs:
        return
    await log(clean(nbs))
    nbs = get_nbs(repo, paths)
    if not nbs:
        return
    await log(execute(nbs, paths))
    await log(export(nbs, paths))
    await log(report(nbs, paths))
    commit(repo, paths)


async def log(coroutine):
    """Log the start and end of a coroutine."""
    logger.info(f"<yellow>START</yellow> {coroutine.__name__}")
    await coroutine
    logger.info(f"<green>FINISH</green> {coroutine.__name__}")


# * -------------------------------------------------------------------------------- * #
# * NOTEBOOK PROCESSING


def get_nbs(repo: Repo | None, paths) -> list[str]:
    """Get all notebooks or just the modified ones."""
    return (
        fold_modified_nbs(repo, paths.notebooks)
        if repo
        else fold_docs_nbs(list(paths.notebooks.glob("**/*.ipynb")), paths.notebooks)
    )


async def clean(nbs: list[str]):
    """Clean notebooks."""
    async with TaskGroup() as tg:
        for nb in nbs:
            tg.create_task(clean_notebook(nb))


async def execute(nbs: list[str], parameters: dict[str, Any] | None = None):
    """Execute notebooks."""
    with ProcessPoolExecutor() as executor:
        for nb in nbs:
            executor.submit(
                execute_notebook,
                input_path=nb,
                output_path=nb,
                remove_tagged_cells=["ploomber-engine-error-cell"],
                parameters=parameters or {},
            )
    async with TaskGroup() as tg:
        for nb in nbs:
            tg.create_task(clean_notebook(nb))


async def export(nbs: list[str], paths):
    """Export notebooks to Markdown and HTML."""
    async with TaskGroup() as tg:
        for nb in nbs:
            tg.create_task(export_notebook(nb, paths))


async def report(nbs: list[str], paths):
    """Generate DOCX reports."""
    async with TaskGroup() as tg:
        for nb in nbs:
            tg.create_task(
                report_on_notebook(**{
                    kwarg: fold(path)
                    for kwarg, path in dict(
                        workdir=paths.md,
                        template=paths.template,
                        filt=paths.filt,
                        zotero=ZOTERO,
                        csl=paths.csl,
                        docx=paths.docx / Path(nb).with_suffix(".docx").name,
                        md=paths.md / Path(nb).with_suffix(".md").name,
                    ).items()
                })
            )


def commit(repo, paths):
    """Commit changes."""
    docs_dvc_file = fold(paths.notebooks.with_suffix(".dvc"))
    repo.commit(docs_dvc_file, force=True)
    add(repo, docs_dvc_file)


# * -------------------------------------------------------------------------------- * #
# * SINGLE NOTEBOOK PROCESSING


async def clean_notebook(nb: str):
    """Clean a notebook."""
    commands = [
        f"ruff --fix-only {nb}",
        f"ruff format {nb}",
         "nb-clean clean"
         " --remove-empty-cells"
         " --preserve-cell-outputs"
         " --preserve-cell-metadata special tags"
        f" -- {nb}",
    ]  # fmt: skip
    for command in commands:
        await run_process(command)


async def export_notebook(nb: str, paths):
    """Export a notebook to Markdown and HTML."""
    md = fold(paths.md)
    html = fold(paths.html)
    for command in [
        f"jupyter-nbconvert --to markdown --no-input --output-dir {md} {nb}",
        f"jupyter-nbconvert --to html --no-input --output-dir {html} {nb}",
    ]:
        await run_process(command)


def preserve_dir(f: Callable[..., Coroutine[Any, Any, Any]]):
    """Preserve the current directory."""

    @wraps(f)
    async def wrapped(*args, **kwargs):
        return await CoroWrapper(f(*args, **kwargs), PreserveDir())

    return wrapped


@preserve_dir
async def report_on_notebook(
    workdir: str, template: str, filt: str, zotero: str, csl: str, docx: str, md: str
):
    """Generate a DOCX report from a notebook.

    Requires changing the active directory to the Markdown folder outside of this
    asynchronous function, due to how Pandoc generates links inside the documents.
    """
    chdir(workdir)
    await run_process(
        venv=False,
        command=(
            " pandoc"
            # Pandoc configuration
            "   --standalone"  # Don't produce a document fragment.
            "   --from markdown-auto_identifiers"  # Avoids bookmark pollution around Markdown headers
            "   --to docx"  # The output format
            f"  --reference-doc {template}"  # The template to export literature reviews to
            # Custom filter to strip out dataframes
            f"  --filter {filt}"
            # Zotero Lua filter and metadata passed to it
            f"  --lua-filter {zotero}"  # Needs to be the one downloaded from the tutorial page https://retorque.re/zotero-better-bibtex/exporting/pandoc/#from-markdown-to-zotero-live-citations
            "   --metadata zotero_library:3"  # Corresponds to "Nucleate pool boiling [3]"
            f"  --metadata zotero_csl_style:{csl}"  # Must also be installed in Zotero
            # I/O
            f"  --output {docx}"
            f"  {md}"
        ),
    )


# * -------------------------------------------------------------------------------- * #
# * UTILITIES

COLORS = {
    "jupyter-nbconvert": "blue",
    "nb-clean": "magenta",
    "ruff": "cyan",
    "pandoc": "red",
}


async def run_process(command: str, venv: bool = True):
    """Run a subprocess asynchronously."""
    c, *args = split(command)
    file = args[-1].split("/")[-1]
    colored_command = f"<{COLORS[c]}>{c}</{COLORS[c]}>"
    logger.info(f"    <yellow>Start </yellow> {colored_command} {file}")
    message = await notebooks.run_process(command, venv)
    logger.info(
        f"    <green>Finish</green> {colored_command} {file}"
        + ((": " + message.replace("\n", ". ")[:30] + "...") if message else "")
    )


def fold_modified_nbs(repo: Repo, docs: Path) -> list[str]:
    """Fold the paths of modified documentation notebooks."""
    modified = get_modified_files(repo)
    return (
        fold_docs_nbs(modified, docs)
        if any(path.is_relative_to(docs) for path in modified)
        else []
    )


def get_modified_files(repo: Repo) -> list[Path]:
    """Get files considered modified by DVC."""
    status = repo.data_status(granular=True)
    modified: list[Path] = []
    for key in ["modified", "added"]:
        paths = status["uncommitted"].get(key) or []
        paths.extend(status["committed"].get(key) or [])
        modified.extend([Path(path).resolve() for path in paths])
    return modified


def fold_docs_nbs(paths: list[Path], docs: Path) -> list[str]:
    """Fold the paths of documentation-related notebooks."""
    return list({
        fold(nb)
        for nb in sorted({
            path.with_suffix(".ipynb")
            for path in paths
            # Consider notebook modified even if only its `.h5` file is
            if path.is_relative_to(docs) and path.suffix in [".ipynb", ".h5"]
        })
    })


# * -------------------------------------------------------------------------------- * #
# * PRIMITIVES


class PreserveDir:
    """Re-entrant context manager that preserves the current directory.

    Reference: <https://stackoverflow.com/a/64395754/20430423>
    """

    def __init__(self):
        self.outer_dir = Path.cwd()
        self.inner_dir = None

    def __enter__(self):
        self.outer_dir = Path.cwd()
        if self.inner_dir is not None:
            chdir(self.inner_dir)

    def __exit__(self, *exception_info_):
        self.inner_dir = Path.cwd()
        chdir(self.outer_dir)


class CoroWrapper:
    """Wrap `target` to have every send issued in a `context`.

    Reference: <https://stackoverflow.com/a/56079900/1600898>
    """

    def __init__(
        self, target: Coroutine[Any, Any, Any], context: AbstractContextManager[Any]
    ):
        self.target = target
        self.context = context

    # wrap an iterator for use with 'await'
    def __await__(self):
        # unwrap the underlying iterator
        target_iter = self.target.__await__()
        # emulate 'yield from'
        iter_send, iter_throw = target_iter.send, target_iter.throw
        send, message = iter_send, None
        while True:
            # communicate with the target coroutine
            try:
                with self.context:
                    signal = send(message)  # type: ignore  # pyright: 1.1.311
            except StopIteration as err:
                return err.value
            else:
                send = iter_send
            # communicate with the ambient event loop
            try:
                message = yield signal
            except BaseException as err:  # noqa: BLE001
                send, message = iter_throw, err



"""Geometry."""

import numpy as np
from pydantic.v1 import BaseModel, Field, validator

from boilercore.types import Coupon, Rod


class Geometry(BaseModel):
    """The fixed geometry for the problem."""

    # Prefix with underscore to exclude from schema
    _in_p_m: float = 39.3701  # (in/m) Conversion factor

    # ! DIAMETER

    diameter: float = Field(
        default=0.375, description="The common diameter of all rods."
    )

    @validator("diameter", always=True)
    @classmethod
    def validate_diameter(cls, diameter):
        """Convert from inches to meters."""
        return diameter / cls._in_p_m

    # ! RODS

    rods: dict[Rod, list[float]] = Field(
        default={
            "X": [3.5253, 3.0500, 2.5756, 2.1006, 0.3754],
            "Y": [3.5250, 3.0504, 2.5752, 2.1008, 0.3752],
            "R": [4.1000, 3.6250, 3.1500, 2.6750, 0.9500],
            "W": [3.5250, 3.0500, 2.5750, 2.1000, 0.3750],
        },
        description="Distance of each thermocouple from the cool side of the rod, starting with TC1. Fifth thermocouple may be omitted. Input: inch. Output: meter.",
    )

    @validator("rods", pre=True, always=True)
    @classmethod
    def validate_rods(cls, rods):
        """Convert from inches to meters."""
        return {
            rod: list(np.array(values) / cls._in_p_m) for rod, values in rods.items()
        }

    # ! COUPONS

    coupons: dict[Coupon, float] = Field(
        default={
            "A0": 0.000,
            "A1": 0.766,
            "A2": 0.770,
            "A3": 0.769,
            "A4": 0.746,
            "A5": 0.734,
            "A6": 0.750,
            "A7": 0.753,
            "A8": 0.753,
            "A9": 0.553,
        },
        description="Length of the coupon. Input: inch. Output: meter.",
    )

    @validator("coupons", pre=True, always=True)
    @classmethod
    def validate_coupons(cls, coupons):
        """Convert from inches to meters."""
        return {coupon: value / cls._in_p_m for coupon, value in coupons.items()}


GEOMETRY = Geometry()



"""Basic models."""

from collections.abc import Callable, Mapping, Sequence
from pathlib import Path
from typing import Any, get_origin

from pydantic.v1 import BaseModel, validator
from ruamel.yaml import YAML

from boilercore.models.types import PathOrPaths, Paths

YAML_INDENT = 2
yaml = YAML()
yaml.indent(mapping=YAML_INDENT, sequence=YAML_INDENT, offset=YAML_INDENT)
yaml.width = 1000  # Otherwise Ruamel breaks lines illegally
yaml.preserve_quotes = True


class YamlModel(BaseModel):
    """Model of a YAML file with automatic schema generation.

    Updates a JSON schema next to the YAML file with each initialization.
    """

    def __init__(self, data_file: Path, **kwargs):
        """Initialize and update the schema."""
        self.update_schema(data_file)
        params = self.get_params(data_file)
        super().__init__(**(params | kwargs))

    def get_params(self, data_file: Path) -> dict[str, Any]:
        """Get parameters from file."""
        return yaml.load(data_file) or {} if data_file.exists() else {}

    def update_schema(self, data_file: Path):
        """Update the schema file next to the data file."""
        schema_file = data_file.with_name(f"{data_file.stem}_schema.json")
        schema_file.write_text(
            encoding="utf-8", data=f"{self.schema_json(indent=YAML_INDENT)}\n"
        )


class SynchronizedPathsYamlModel(YamlModel):
    """Model of a YAML file that synchronizes paths back to the file.

    For example, synchronize complex path structures back to `params.yaml` DVC files for
    pipeline orchestration.
    """

    def __init__(self, data_file: Path, **kwargs):
        """Initialize and update the schema."""
        super().__init__(data_file, **kwargs)

    def get_params(self, data_file: Path) -> dict[str, Paths[Path]]:
        """Get parameters from file, synchronizing paths in the file."""
        params = super().get_params(data_file)
        paths = self.get_paths()
        yaml.dump(params | paths, data_file)
        for i, param in paths.items():
            for j, p in param.items():
                paths[i][j] = apply_to_path_or_paths(p, lambda p_: Path(p_).resolve())
        return params | paths

    def get_paths(self) -> dict[str, Paths[str]]:
        """Get all paths specified in paths-type models."""
        maybe_excludes = self.__exclude_fields__
        excludes = set(maybe_excludes.keys()) if maybe_excludes else set()
        defaults: dict[str, Paths[str]] = {}
        for key, field in self.__fields__.items():
            if key in excludes:
                continue
            if generic_ := get_origin(field.type_):
                type_ = type(generic_)
            else:
                type_ = field.type_
            if issubclass(type_, DefaultPathsModel):
                defaults[key] = type_.get_paths()
        return defaults


def check_pathlike(model: BaseModel, field: str, type_: type):
    """Check that the field is path-like."""
    if not issubclass(type_, Path):
        raise TypeError(
            f"Field <{field}> is not Path-like in {model}, derived from {DefaultPathsModel}."
        )


class DefaultPathsModel(BaseModel):
    """All fields must be path-like and have defaults specified in this model."""

    class Config:
        """Model config."""

        @staticmethod
        def schema_extra(schema: dict[str, Any], model):
            """Replace backslashes with forward slashes in paths."""
            if schema.get("required"):
                raise TypeError(
                    f"Defaults must be specified in {model}, derived from {DefaultPathsModel}."
                )
            for (field, prop), type_ in zip(
                schema["properties"].items(),
                (field.type_ for field in model.__fields__.values()),
                strict=True,
            ):
                # If default is a container, `type_` will be the type of its elements.
                check_pathlike(model, field, type_)
                prop["default"] = apply_to_path_or_paths(prop.get("default"), pathfold)

    @classmethod
    def get_paths(cls) -> Paths[str]:
        """Get the paths for this model."""
        return {
            key: value["default"] for key, value in cls.schema()["properties"].items()
        }


def apply_to_path_or_paths(
    path_or_paths: PathOrPaths[Any], fun: Callable[[Any], Any]
) -> PathOrPaths[Any]:
    """Apply a function to a path, sequence of paths, or mapping of names to paths."""
    if isinstance(path_or_paths, str):
        return fun(path_or_paths)
    elif isinstance(path_or_paths, Sequence):
        return [fun(path) for path in path_or_paths]
    elif isinstance(path_or_paths, Mapping):
        return {key: fun(path) for key, path in path_or_paths.items()}
    else:
        raise TypeError("Type not supported.")


class CreatePathsModel(DefaultPathsModel):
    """Parent directories will be created for all fields in this model."""

    @validator("*", always=True, pre=True, each_item=True)
    @classmethod
    def create_directories(cls, value):
        """Create directories associated with each value."""
        path = Path(value)
        if path.is_file():
            return value
        directory = path.parent if path.suffix else path
        directory.mkdir(parents=True, exist_ok=True)
        return value


def pathfold(path: str) -> str:
    """Return the shortest possible path with forward slashes."""
    return str(prefer_relative_path(Path(path))).replace("\\", "/")


def prefer_relative_path(path: Path) -> Path:
    """Return the path relative to the current working directory if possible."""
    cwd = Path.cwd()
    return path.relative_to(cwd) if path.is_relative_to(cwd) else path



"""Hash utilities."""

from collections.abc import Callable, Hashable, ItemsView, Iterable, Mapping
from inspect import getsource, signature
from typing import Any

from cachier.config import _default_hash_func

from boilercore.types import Freezable


def hash_args(
    fun: Callable[..., Any], args: Iterable[Any], kwds: Mapping[str, Any]
) -> str:
    """Hash a particular set of arguments meant to be passed to a particular function.

    Passing this function, with `fun` applied via `partial`, to `cachier`'s `hash_func`
    decorator allows us to create a cacheable form of `fun`.

    Returns: Hash of the arguments.

    Args:
        fun: Function from which the signature will be generated.
        args: Positional arguments to hash.
        kwds: Keyword arguments to hash.

    Example:
        ```Python
        import cachier
        from functools import partial

        @cachier(hash_func=partial(hash_args_PATCHED, uncached_function))
        def cached_function(*args, *kwds):
            return uncached_function(*args, *kwds)
        ```
    """
    bound_args = signature(fun).bind(*args, **kwds).arguments
    return _default_hash_func(
        (), {param: freeze(val) for param, val in bound_args.items()}
    )


def freeze(v: Hashable | Freezable) -> Hashable:
    """Make value hashable."""
    match v:
        case Hashable():
            return v
        case Callable():  # type: ignore  # pyright: 1.1.347
            getsource(v)
        case Mapping():
            return tuple((k, freeze(v)) for k, v in v.items())
        case ItemsView():
            return tuple((k, freeze(v)) for k, v in v)
        case Iterable():
            return tuple(freeze(v) for v in v)
        case _:
            raise TypeError(
                f"{type(v)} not Hashable, Callable, Mapping, ItemsView, or Iterable."
            )



"""Types."""

from collections.abc import Callable, Iterable, Mapping
from typing import Any, TypeAlias

Params: TypeAlias = Mapping[str, Any]
Attributes: TypeAlias = Iterable[str]

SimpleNamespaceReceiver: TypeAlias = Callable[..., Any]
"""Should be a `Callable` with an `ns` parameter expecting a `SimpleNamespace`."""



"""Project paths."""

from pydantic.v1 import DirectoryPath, FilePath

import boilercore
from boilercore import PROJECT_PATH
from boilercore.models import CreatePathsModel
from boilercore.paths import get_package_dir, map_stages


class Paths(CreatePathsModel):
    """Paths relevant to the project."""

    # * Roots
    project: DirectoryPath = PROJECT_PATH
    package: DirectoryPath = get_package_dir(boilercore)
    data: DirectoryPath = project / "data"
    stages: dict[str, FilePath] = map_stages(
        package / "stages", suffixes=[".py", ".ipynb"]
    )

    # * DVC-tracked inputs
    fit: FilePath = package / "models/fit.py"
    # ! Scripts
    scripts: DirectoryPath = data / "scripts"
    filt: FilePath = scripts / "filt.py"
    csl: FilePath = scripts / "international-journal-of-heat-and-mass-transfer.csl"
    template: FilePath = scripts / "template.dotx"

    # * DVC-tracked results
    models: DirectoryPath = data / "models"



"""Common functionality of boiler repositories."""

from collections.abc import Sequence
from importlib.machinery import ModuleSpec
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import NamedTuple
from warnings import filterwarnings

from boilercore.paths import get_module_name
from boilercore.types import Action

PROJECT_PATH = Path()


def get_params_file():  # noqa: D103
    return PROJECT_PATH / "params.yaml"


class WarningFilter(NamedTuple):
    """A warning filter, e.g. to be unpacked into `warnings.filterwarnings`."""

    action: Action = "ignore"
    message: str = ""
    category: type[Warning] = Warning
    module: str = ""
    lineno: int = 0
    append: bool = False


DEFAULT_CATEGORIES = [DeprecationWarning, PendingDeprecationWarning, EncodingWarning]
ERROR = "error"
DEFAULT = "default"
NO_WARNINGS = []


def filter_certain_warnings(
    package: ModuleType | ModuleSpec | Path | str,
    categories: Sequence[type[Warning]] = DEFAULT_CATEGORIES,
    root_action: Action | None = ERROR,
    package_action: Action = ERROR,
    other_action: Action = DEFAULT,
    other_warnings: Sequence[WarningFilter] = NO_WARNINGS,
):
    """Filter certain warnings for a package."""
    for filt in [
        # Optionally filter warnings with the root action
        *([WarningFilter(action=root_action)] if root_action else []),
        # Filter certain categories with a package action, and third-party action otherwise
        *chain.from_iterable(
            filter_package_warnings(
                package=package,
                category=category,
                action=package_action,
                other_action=other_action,
            )
            for category in categories
        ),
        # Ignore this as it crops up only during test time under some configurations
        WarningFilter(
            message=r"ImportDenier\.find_spec\(\) not found; falling back to find_module\(\)",
            category=ImportWarning,
        ),
        # Additionally filter these other warnings
        *other_warnings,
    ]:
        filterwarnings(*filt)


def filter_package_warnings(
    package: ModuleType | ModuleSpec | Path | str,
    category: type[Warning],
    action: Action = ERROR,
    other_action: Action = DEFAULT,
) -> tuple[WarningFilter, WarningFilter]:
    """Get filter which filters warnings differently for the package."""
    all_package_modules = rf"{get_module_name(package)}\..*"
    return (
        WarningFilter(action=other_action, category=category),
        WarningFilter(action=action, category=category, module=all_package_modules),
    )



"""Symbolic parameter groups for model fitting."""

from sympy import Function, symbols

from boilercore.models.fit import FIT

params = symbols(["x", *FIT.free_params, *FIT.fixed_params])
(x, T_s, q_s, h_a, k, h_w) = params

inputs = symbols(list(FIT.model_inputs.keys()))
(r, T_infa, T_infw, x_s, x_wa) = inputs

intermediate_vars = symbols(
    """
    h,
    q_0,
    q_wa,
    T_0,
    T_inf,
    T_wa,
    x_0,
    """
)
(
    h,  # (W/m^2-K) Convection heat transfer coefficient
    q_0,  # (W/m^2) q at x_0, the LHS of a general domain
    q_wa,  # (W/m^2) q at the domain interface
    T_0,  # (C) T at x_0, the LHS of a general domain
    T_inf,  # (C) Ambient temperature
    T_wa,  # (C) T at the domain interface
    x_0,  # (m) x at the LHS of a general domain
) = intermediate_vars

functions = symbols(
    """
    T*,
    T_a,
    T_w,
    T,
    """,
    cls=Function,  # type: ignore  # sympy
)
(
    T_int,  # (T*, C) The general solution to the ODE
    T_a,  # (C) Solution in air
    T_w,  # (C) Solution in water
    T,  # (C) The piecewise combination of the two above solutions
) = functions



"""Model function generation tools."""

import warnings
from collections.abc import Callable
from functools import wraps
from pathlib import Path
from sys import version_info
from typing import Any

import dill
import numpy as np


def get_model(models: Path):
    """Unpickle the model function for fitting data."""
    file_bytes = Path(
        models
        / f"modelfun-{'.'.join([str(elem) for elem in version_info[:2]])}.dillpickle"
    ).read_bytes()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", dill.UnpicklingWarning)
        unpickled_model = dill.loads(file_bytes)
    return unpickled_model.basic, fix_model(unpickled_model.for_ufloat)


def fix_model(f) -> Callable[..., Any]:
    """Fix edge-cases of lambdify where all inputs must be arrays.

    See the notes section in the link below where it says, "However, in some cases
    the generated function relies on the input being a numpy array."

    https://docs.sympy.org/latest/modules/utilities/lambdify.html#sympy.utilities.lambdify.lambdify
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(
            *(np.array(arg) for arg in args),
            **{k: np.array(v) for k, v in kwargs.items()},
        )

        return result if result.size > 1 else result.item()

    return wrapper



"""Test helpers."""

from collections.abc import Callable
from pathlib import Path
from shutil import copytree
from types import ModuleType
from typing import Any, NamedTuple

import pytest
from _pytest.python import Function
from numpy.typing import ArrayLike


class MFParam(NamedTuple):
    """Parameter for model function tests."""

    run: str
    y: ArrayLike
    expected: dict[str, float]


def get_session_path(
    tmp_path_factory: pytest.TempPathFactory, package: ModuleType
) -> Path:
    """Copy test data to a session path and return the path."""
    test_data_name = Path("root")
    project_test_data = Path("tests") / test_data_name
    session_path = tmp_path_factory.getbasetemp() / test_data_name
    package.PROJECT_PATH = session_path  # type: ignore
    copytree(project_test_data, session_path, dirs_exist_ok=True)
    return session_path


def unwrap_node(node: Function) -> Callable[..., Any]:
    """Unwrap a pytest node."""
    return getattr(node.module, node.originalname)



"""Fit parameters."""

import numpy as np
from pydantic.v1 import BaseModel, Field, validator

from boilercore.models.types import Bound

FIT_METHOD = "trf"
"""Default curve fitting method. Supports bounded curve fits."""

MIN_NONZERO = 1e-3
"""Smallest nonzero boundary or initial guess to substitute for zero."""

INIT_H = 1.0
"""Initial heat transfer coefficient. The model is sensitive to very low h."""

MODEL_PARAMS = ["T_s", "q_s", "k", "h_a", "h_w"]
"""Model parameters."""

FIXED_PARAMS = ["k", "h_w"]
"""Fixed parameters in the model."""


def get_model_errors(params: list[str]) -> list[str]:
    """Get error parameters for model parameters."""
    return [f"{param}_err" for param in params]


class Fit(BaseModel):
    """Parameters for model fit."""

    fit_method = "trf"
    """Model fit method."""
    model_params = MODEL_PARAMS
    """Parameters that can vary in the model. Some will be fixed."""
    fixed_params = FIXED_PARAMS
    """Parameters to fix. Evaluated before fitting, overridable in code."""
    free_params = [param for param in MODEL_PARAMS if param not in FIXED_PARAMS]
    """Free parameters."""
    free_errors = get_model_errors(free_params)
    model_errors = get_model_errors(model_params)
    fixed_errors = get_model_errors(fixed_params)
    params_and_errors = [*model_params, *model_errors]

    model_inputs: dict[str, float] = Field(
        default=dict(
            r=0.0047625,  # (m)
            T_infa=25.0,  # (C)
            T_infw=100.0,  # (C)
            x_s=0.0,  # (m)
            x_wa=0.0381,  # (m)
        ),
        description="Inputs to the symbolic model float evaluation stage.",
    )

    # ! MODEL BOUNDS

    model_bounds: dict[str, tuple[Bound, Bound]] = Field(
        default=dict(
            T_s=(-273.0, "inf"),  # (C) T_s
            q_s=("-inf", "inf"),  # (W/m^2) q_s
            k=(MIN_NONZERO, "inf"),  # (W/m-K) k
            h_a=(MIN_NONZERO, "inf"),  # (W/m^2-K) h_a
            h_w=(MIN_NONZERO, "inf"),  # (W/m^2-K) h_w
        ),
        description="Bounds for the model parameters. Not used if parameter is fixed.",
    )

    @validator("model_bounds", always=True)
    @classmethod
    def validate_model_bounds(cls, model_bounds) -> dict[str, tuple[float, float]]:
        """Substitute np.inf for 'inf' and avoid exact zero lower bounds."""
        for param, b in model_bounds.items():
            if isinstance(b[0], str) and "-inf" in b[0]:
                b0 = -np.inf
            elif b[0] == 0.0:
                b0 = MIN_NONZERO
            else:
                b0 = b[0]
            b1 = np.inf if isinstance(b[0], str) and "inf" in b[1] else b[1]
            model_bounds[param] = (b0, b1)
        return model_bounds

    # ! INITIAL VALUES

    initial_values: dict[str, float] = Field(
        default=dict(
            T_s=100.0,  # (C)
            q_s=10.0,  # (W/cm^2)
            k=400.0,  # (W/m-K)
            h_a=INIT_H,  # (W/m^2-K)
            h_w=INIT_H,  # (W/m^2-K)
        ),
        description="Initial guess for free parameters, constant value otherwise.",
    )

    @validator("initial_values", always=True)
    @classmethod
    def validate_initial_values(cls, model_inputs) -> dict[str, float]:
        """Avoid exact zero guesses."""
        return {
            param: MIN_NONZERO if v == 0.0 else v for param, v in model_inputs.items()
        }

    @property
    def fixed_values(self) -> dict[str, float]:
        """Fixed parameter values."""
        return {k: v for k, v in self.initial_values.items() if k in self.fixed_params}


FIT = Fit()
"""Fit parameters."""



"""Trials."""

import datetime
import re
from pathlib import Path

import numpy as np
import pandas as pd
from pydantic.v1 import BaseModel, DirectoryPath, Field, FilePath, validator

from boilercore.models import YamlModel
from boilercore.models.geometry import Geometry
from boilercore.types import Coupon, Group, Joint, Rod, Sample


class Trial(BaseModel):
    """A trial."""

    # ! META FIELDS ADDED TO DATAFRAME

    # !! DATE

    date: datetime.date = Field(
        default=..., description="The date of the trial.", exclude=True
    )

    @validator("date", pre=True)
    @classmethod
    def validate_date(cls, date):
        """Validate a date."""
        return datetime.date.fromisoformat(date)

    group: Group
    rod: Rod
    coupon: Coupon
    sample: Sample | None
    joint: Joint
    good: bool = Field(default=True, description="Whether the boiling curve is good.")
    plot: bool = Field(
        default=False, description="Whether this trial should be plotted."
    )

    # ! FIELDS TO EXCLUDE FROM DATAFRAME

    @property
    def timestamp(self):
        """Trial timestamp."""
        return pd.Timestamp(self.date)

    # Loaded from config, but not propagated to dataframes. Not readable in a table
    # anyways, and NA-handling results in additional ~40s to pipeline due to the need to
    # use slow "fillna".
    comment: str = Field(default="", exclude=True)

    # ! PROJECT-DEPENDENT SETUP

    # Can't be None. Set in Project.__init__()
    path: DirectoryPath = Field(default=None, exclude=True)
    run_files: list[FilePath] = Field(default=None, exclude=True)
    run_index: list[tuple[pd.Timestamp, pd.Timestamp]] = Field(
        default=None, exclude=True
    )
    thermocouple_pos: dict[str, float] = Field(default=None, exclude=True)

    def setup(self, paths, geometry: Geometry, copper_temps: list[str]):
        """Set up the trial."""
        self.set_paths(paths)
        self.set_geometry(geometry, copper_temps)

    def set_paths(self, paths):
        """Get the path to the data for this trial. Called during project setup."""
        trial_base = paths.trials / self.date.isoformat()
        self.path = trial_base
        self.run_files = sorted(self.path.glob("*.csv"))
        if not self.run_files:
            raise FileNotFoundError(f"No runs found in {self.path}.")
        self.set_index()

    def set_index(self):
        """Get the multiindex for all runs. Called during project setup."""
        run_re = re.compile(r"(?P<date>.*)T(?P<time>.*)")
        run_index: list[tuple[pd.Timestamp, pd.Timestamp]] = []
        for run_file in self.run_files:
            run_time = run_file.stem.removeprefix("results_")

            if m := run_re.match(run_time):
                run_time = f"{m['date']}T{m['time'].replace('-', ':')}"
            else:
                raise AttributeError(f"Could not parse run time: {run_time}")

            trial_date = self.date.isoformat()  # for input to fromisoformat() below
            run_index.append(
                tuple(  # type: ignore  # It's indeterminate but we hard-code the inputs  # pyright 1.1.325
                    pd.Timestamp.fromisoformat(item) for item in (trial_date, run_time)
                )
            )
        self.run_index = run_index

    def set_geometry(self, geometry: Geometry, copper_temps: list[str]):
        """Get relevant geometry for the trial."""
        thermocouple_pos = list(
            np.array(geometry.rods[self.rod]) + np.array(geometry.coupons[self.coupon])
        )
        self.thermocouple_pos = dict(zip(copper_temps, thermocouple_pos, strict=True))


class Trials(YamlModel):
    """The trials."""

    trials: list[Trial]

    def __init__(self, data_file: Path):
        super().__init__(data_file)



"""Types."""

from collections.abc import Callable, ItemsView, Iterable, Mapping
from typing import Any, Literal, TypeAlias, TypeVar

Bound = TypeVar("Bound", bound=tuple[float | str, float | str])
"""Boundary for a parameter to be fitted."""

Guess = TypeVar("Guess", bound=float)
"""Guess for a parameter to be fitted."""

Coupon = Literal["A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8", "A9"]
"""The coupon attached to the rod for this trial."""

Rod = Literal["W", "X", "Y", "R"]
"""The rod used in this trial."""

Group = Literal["control", "porous", "hybrid"]
"""The group that this sample belongs to."""

Joint = Literal["paste", "epoxy", "solder", "none"]
"""The method used to join parts of the sample in this trial."""

Sample = Literal["B3"]
"""The sample attached to the coupon in this trial."""

Action: TypeAlias = Literal["default", "error", "ignore", "always", "module", "once"]
"""Action to take for a warning."""
Freezable: TypeAlias = (
    Callable[..., Any] | Mapping[str, Any] | ItemsView[str, Any] | Iterable[Any]
)
"""Value that can be frozen."""
